@@ -9,3 +9,6 @@
 #undef HAVE_STPCPY
 #undef GNOME_EXPLICIT_TRANSLATION_DOMAIN
 #undef GETTEXT_PACKAGE
+#undef HAVE_BIND_TEXTDOMAIN_CODESET
+#undef HAVE_RENDER
+#define HAVE_GTK_MULTIHEAD /* needed for egg-screen-exec.c */
